import gitlab
import logging
import urllib3
import json
from re import search

urllib3.disable_warnings()
logger = logging.getLogger("gitlab_client")
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s %(levelname)s - %(message)s")
handler_console = logging.StreamHandler()
handler_console.setFormatter(formatter)
logger.addHandler(handler_console)

class GitLabClient():

    def __init__(self, token, project_id, branch, url="https://gitlab.bit9.local"):
        self.url = url
        self.token = token
        self.project_id = project_id
        self.branch = branch
        self.commit_title = 'Created by RBAC-Clean Summoner: {}'.format(self.branch)
        self.connection = gitlab.Gitlab(self.url, self.token, api_version="4", ssl_verify="/etc/ssl/certs/combined.crt")
        self.project = self.connection.projects.get(self.project_id)

    def get_branch_commits(self):
        return self.project.commits.list(ref_name=self.branch)

    def get_merge_request(self, branch):
        '''
        Validate that the merge request exists for a given branch.
        Returns None is the MR does not exist.
        '''
        merge_requests = self.project.mergerequests.list(state='opened', order_by='created_at')
        for merge_request in merge_requests:
             # Detect only MRs into master
            if merge_request.source_branch == branch and merge_request.target_branch == 'master':
                return merge_request
        return None

    def get_watchlist_changed_files(self, mr):
        '''
        Check the watch list items and if the MR diff has changed items in that watchlist, return true
        '''

        watchlist_regex = '\/?db\/secops\/(roles|permissions|subroles)\/.[\w:.-]+.json'

        diff_list = mr.changes()['changes']
        watchlist_changed_files = []

        for diff in diff_list:
            if (search(watchlist_regex, diff['old_path']) or search(watchlist_regex, diff['new_path'])) and not diff['deleted_file']:
                watchlist_changed_files.append(diff['new_path'])
        return watchlist_changed_files

    def get_file(self, file_path):
        '''
        Returns dictionary with file path as key and contents as value
        '''
        file = self.project.files.get(file_path, ref=self.branch).decode().decode('utf-8')

        return file

    def make_commit(self, files):
        # Ensure we never commit to master or develop
        if (self.branch != 'master' and self.branch != 'develop'):
            commit_request = { 'branch': self.branch, 'commit_message': self.commit_title, 'actions': [] }
            files_to_commit = []
            for file in files:
                files_to_commit.append(file['file_path'])
                commit_request['actions'].append(
                    { 'action': 'update',
                      'file_path': file['file_path'],
                      'content': file['content'] })
            logger.info("RBAC files that require formatting:\n\t%s" %
                        '\n\t'.join(str(file) for file in files_to_commit))
            logger.info("Creating commit with request:\n%s" % json.dumps(commit_request, indent=4))
            commit = self.project.commits.create(commit_request)
            logger.info("Commit with SHA %s and title %s created" % (commit.short_id, commit.title))
