from os import environ
import json
import logging
import urllib3
import hashlib
from gitclient import GitLabClient
from sanitizer import sanitize
from sanitizer import sanitize_all_and_save

urllib3.disable_warnings()
logger = logging.getLogger("rbac-clean-summoner")
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s %(levelname)s - %(message)s")
handler_console = logging.StreamHandler()
handler_console.setFormatter(formatter)
logger.addHandler(handler_console)

def main():

    if environ['GITLAB_TOKEN_RBAC_CLEAN'] == environ['LOCAL_RUN_FLAG'] and environ.get('CI_COMMIT_REF_NAME') == environ['LOCAL_RUN_FLAG']:
        logger.info("Running sanitize locally on all RBAC files")
        sanitize_all_and_save()
    else:
        logger.info("Running RBAC Summoner in Pipeline ")

        # Project ID for Access Control Service
        project_id = 859
        token = environ['GITLAB_TOKEN_RBAC_CLEAN']
        branch = environ.get('CI_COMMIT_REF_NAME')

        gl_client = GitLabClient(token, project_id, branch)

        merge_request = gl_client.get_merge_request(gl_client.branch)

        if merge_request is None:
            logger.error("No Merge Request is open for branch %s. Please open MR into master and re-run"
                        % environ.get('CI_COMMIT_REF_NAME'))
            return

        changed_rbac_files = gl_client.get_watchlist_changed_files(merge_request)

        if len(changed_rbac_files) == 0:
            logger.info("No RBAC JSON files changed in branch %s. " % environ.get('CI_COMMIT_REF_NAME'))
        else:
            logger.info("Detected the following RBAC JSON file(s) changed in MR:\n\t%s" %
                    '\n\t'.join(str(file) for file in changed_rbac_files))

        if len(changed_rbac_files) != 0:
            sanitized_files = []
            for file_name in changed_rbac_files:
                file_contents = gl_client.get_file(file_name)

                file_contents_hash = hashlib.md5(file_contents.encode('utf-8')).hexdigest()

                # Update and format the json
                updated_file_dict = { "file_path": file_name,
                                     "content" : json.dumps(sanitize(file_name, json.loads(file_contents)), indent=4, sort_keys=True)}

                updated_file_content_hash =hashlib.md5(updated_file_dict['content'].encode('utf-8')).hexdigest()
                if file_contents_hash != updated_file_content_hash:
                    # This file hasn't been sanitized
                    sanitized_files.append(updated_file_dict)
            if len(sanitized_files) != 0:
                # Make commit with sanitized files
                gl_client.make_commit(sanitized_files)
            else:
                logger.info("No changed files to sanitize.")
    logger.info("Done.")

if __name__ == "__main__":
    main()
