import argparse
import json
import hashlib
import logging
import os
import yaml
import requests
import time
import urllib3

from requests_sessions import *

import rbac_util_encoding

urllib3.disable_warnings()
logger = logging.getLogger("generate_service_connector")
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s %(levelname)s - %(message)s")
handler_console = logging.StreamHandler()
handler_console.setFormatter(formatter)
logger.addHandler(handler_console)


def main():
    connectors_directory = os.path.dirname(os.path.abspath(__file__ + "/.."))

    parser = argparse.ArgumentParser(description="Generate service connectors and associate grants")
    parser.add_argument('-u', "--url",
                        help="Specifiy the base URL to access psc.access instead of using environment lookup tables",
                        dest='base_url', required=True)
    parser.add_argument("--request-context", help="Set the X-Request-Context header to this value - default omitted",
                        dest='request_context', required=True)
    parser.add_argument("--user-context", help="Set the X-User-Context header to this value - default omitted",
                        dest='user_context', required=True)
    parser.add_argument("--org-key", help="Set the org key for org 1", dest='org_key', required=True)
    parser.add_argument("--salt",
                        help="The salt use to encode connectorId and apiKey. If omitted, the org-key will be used as a salt.",
                        dest='salt', required=False)

    args = parser.parse_known_args()[0]

    base_url = args.base_url

    headers = {
        'Content-Type': 'application/json',
        "X-Request-Context": args.request_context,
        "X-User-Context": args.user_context
    }

    requests_session = MigrationSession(base_url)
    requests_session.headers.update(headers)

    org_key = args.org_key

    salt = org_key
    if args.salt:
        salt = salt + args.salt

    for filename in os.listdir("{}/connectors/orgs/1".format(connectors_directory)):
        if filename.endswith(".json"):
            try:
                f = open("{}/connectors/orgs/1/{}".format(connectors_directory, filename), 'r')
            except IOError:
                print("ERROR: An IO error occured when reading from file roles/{}".format(filename))
            else:
                text = f.read()
                service_connector_json = json.loads(text)
                service_connector_json['role_urn'] = service_connector_json['role_urn'].format(org_key=org_key)

                service_connector_name = service_connector_json['connector_name']
                service_connector_urn = service_connector_json['role_urn']
                
                # Actual length is dklen*2
                connector_id = rbac_util_encoding.encode_connector_id(service_connector_name, salt)
                api_key = rbac_util_encoding.encode_api_key(service_connector_name, salt)

                logger.info("----------------------------------------------------------------------------------")

                # Verify Service Connector has correct Service Role
                principle_urn = "psc:cnn:{}:{}".format(org_key, connector_id)
                org_urn = "psc:org:{}".format(org_key)
                grants_url = "/access/v1/grants/{}/org-ref/{}".format(principle_urn, org_urn)

                grant_response = requests_session.get(path=grants_url)

                # Grant exists
                if grant_response.status_code == 200:
                    if service_connector_urn not in grant_response.json()['roles']:
                        logger.info(
                            "Service Connector %s assigned incorrect service role %s. Updating to correct role %s" %
                            (service_connector_name, str(grant_response.json()['roles']), service_connector_urn))
                        grant_update_request = grant_response.json()
                        grant_update_request['roles'] = [service_connector_urn]

                        grants_url = "/access/v1/grants/{}/org-ref/{}".format(principle_urn, org_urn)
                        # Update Service Connector Role
                        grant_update_response = requests_session.put(
                            path=grants_url,
                            data=grant_update_request)
                        grant_update_response.raise_for_status()
                    else:
                        logger.info("Grant for service connector with name '%s' exists" % service_connector_name)
                else:
                    try:
                        principle_urn = "psc:cnn:{}:{}".format(org_key, connector_id)
                        org_urn = "psc:org:{}".format(org_key)
                        logger.info("----------------------------------------------------------------------------------")
                        logger.info("Granting principal '%s' role '%s'" % (principle_urn, service_connector_urn))

                        grants_url = "/access/v1/grants/{}/org-ref/{}".format(principle_urn, org_urn)
                        grant_response = requests_session.post(
                            path=grants_url,
                            data=json.dumps({
                                "principal": principle_urn,
                                "expires": None,
                                "revoked": False,
                                "roles": [service_connector_urn],
                                "org_ref": org_urn
                            })
                        )
                        grant_response.raise_for_status()

                        logger.info(
                            "Grant principal '%s' role '%s' successful" % (principle_urn, service_connector_urn))

                    except requests.HTTPError as ex:
                        logger.error("Request failed (json=%s)" % ex.response.__dict__, exc_info=ex)
                        raise ex


logger.info("Done creating service connector grants")


if __name__ == "__main__":
    main()
