from plans import *
import json
import requests
import time


class Executor:
    def __init__(self, org_key, session, csp_session, migration_plan):
        self.org_key = org_key
        self.csp_session = csp_session
        self.session = session
        self.migration_plan = migration_plan

    def execute_all(self):
        print("---------- EXECUTE ----------")
        if self.csp_session:
            for plan in self.migration_plan.permission_plan_csp:
                self.execute_permission_plan_csp(plan)

            for plan in self.migration_plan.subrole_plan_csp:
                self.execute_subrole_plan_csp(plan)

        for plan in self.migration_plan.permission_plan:
            self.execute_permission_plan(plan)

        for plan in self.migration_plan.subrole_plan:
            self.execute_subrole_plan(plan)

        for plan in self.migration_plan.role_plan:
            self.execute_role_plan(plan)

        for plan in self.migration_plan.org_role_plan:
            self.execute_role_plan(plan, self.org_key)

    def execute_permission_plan(self, permission_plan):
        if permission_plan.entity_action == "CREATE":
            self.add_permission(permission_plan.entity_json)
        elif permission_plan.entity_action == "UPDATE":
            self.update_permission(permission_plan.entity_json)

    def execute_permission_plan_csp(self, permission_plan):
        if permission_plan.entity_action == "CREATE":
            self.add_permission_csp(permission_plan.entity_json)
        elif permission_plan.entity_action == "UPDATE":
            self.update_permission_csp(permission_plan.entity_json)

    def execute_subrole_plan(self, subrole_plan):
        if subrole_plan.entity_action == "CREATE":
            self.add_subrole(subrole_plan.entity_json)
        elif subrole_plan.entity_action == "UPDATE":
            self.update_subrole(subrole_plan.entity_json)

        if subrole_plan.permissions_action == "CREATE" \
                or subrole_plan.permissions_action == "UPDATE":
            self.update_subrole_permissions(subrole_plan.entity_json)

    def execute_subrole_plan_csp(self, subrole_plan):
        if subrole_plan.entity_action == "CREATE":
            self.add_subrole_csp(subrole_plan.entity_json)
        elif subrole_plan.entity_action == "UPDATE":
            self.update_subrole_csp(subrole_plan.entity_json)    

    def execute_role_plan(self, role_plan, org_key=None):
        # Update or create the role definition
        if role_plan.entity_action == "CREATE":
            self.add_role(role_plan.entity_json, org_key)
        elif role_plan.entity_action == "UPDATE":
            self.update_role(role_plan.entity_json, org_key)


        # update the role subroles or permissions
        update_permissions = role_plan.permissions_action == "CREATE" \
                or role_plan.permissions_action == "UPDATE"
        update_subroles = role_plan.subroles_action == "CREATE" \
                or role_plan.subroles_action == "UPDATE"

        if update_permissions and update_subroles:
            # In the event that a role is going from subroles to permissions, the subroles must be removed
            # before the permissions can be added (or vice-versa)
            new_subroles = role_plan.entity_json['subroles']
            if len(new_subroles) > 0:
                self.update_role_permissions(role_plan.entity_json, org_key)
                self.update_role_subroles(role_plan.entity_json, org_key)
            else:
                self.update_role_subroles(role_plan.entity_json, org_key)
                self.update_role_permissions(role_plan.entity_json, org_key)
        elif update_permissions:
            self.update_role_permissions(role_plan.entity_json, org_key)
        elif update_subroles:
            self.update_role_subroles(role_plan.entity_json, org_key)

        """ If either the permissions or subroles are updated, we need to
        invalidate the cache entry for that role to prevent their old permission
        set being used until the cache entry expires"""
        if update_permissions or update_subroles:
            self.invalidate_role_cache(role_plan.entity_json)

        # update the role overrides
        if role_plan.overrides_action == "CREATE" \
                or role_plan.overrides_action == "UPDATE":
            self.update_role_overrides(role_plan.entity_json, org_key)

    def add_permission(self, permission):
        """Create a new permission definition"""
        permission_urn = permission['urn']
        post_resp = self.session.post('/access/v2/permissions', data=json.dumps(permission))

        if not post_resp.ok:
            raise RuntimeError('ERROR: could not add permission {}:  {}/{}'.format(permission_urn, post_resp.status_code, post_resp.text))
        else:
            print('Successfully added permission {}'.format(permission_urn))
        
    def add_permission_csp(self, permission):
        """Create a CSP permission"""
        """The name of the permission gets the serviceTicker prepended, once POSTed to CSP"""
        csp_permission_name = permission['urn']
        csp_payload = { "name": csp_permission_name, 
                        "displayName": permission['displayName'], 
                        "description": permission['description'], 
                        "customRoleSupport": "true" }
        retry = True
        while retry:
            post_resp = self.csp_session.post('/csp/gateway/iam-roles-mgmt/api/services/{}/permissions'.format(self.csp_session.service_definition_id), 
                                                                                                           data=json.dumps(csp_payload))
            if post_resp.status_code == 429:
                time.sleep(60)
            else:
                retry = False

        if not post_resp.ok:
            raise RuntimeError('ERROR: could not add CSP permission {}: {}/{}'.format(csp_permission_name, post_resp.status_code, post_resp.text))
        else:
            print("Successfully added permission {} to CSP".format(csp_permission_name))

    def update_permission(self, permission):
        """Update a permission definition"""
        urn = permission['urn']
        # Get the existing permission using the API w/ the URN (optimistic locking)
        existing_resp = self.session.get('/access/v2/permissions/{}'.format(urn))
        if not existing_resp.ok:
            raise RuntimeError('ERROR: could not retrieve permission {} to update: {}/{}'.format(urn, existing_resp.status_code, existing_resp.text))
        existing_permission = json.loads(existing_resp.text)
        permission['version'] = existing_permission['version']

        put_resp = self.session.put('/access/v2/permissions/{}'.format(urn), data=json.dumps(permission))
        if not put_resp.ok:
            raise RuntimeError('ERROR: could not update permission {}: {}/{}'
                               .format(urn, put_resp.status_code, put_resp.text))
        else:
            print('Successfully updated permission {}'.format(urn))

    def update_permission_csp(self, permission):
        """Update CSP permission"""
        full_csp_permission_name = "{}:{}".format(self.csp_session.service_ticker, permission['urn'])
        api = '/csp/gateway/iam-roles-mgmt/api/services/{}/permissions/{}'.format(self.csp_session.service_definition_id, full_csp_permission_name)
        patch_payload = { "displayName": permission['displayName'], "description": permission['description'] }
        retry = True
        while retry:
            patch_resp = self.csp_session.patch(api, data=json.dumps(patch_payload)) 
            if patch_resp.status_code == 429:
                time.sleep(60)
            else:
                retry = False

        if not patch_resp.ok:
            raise RuntimeError('ERROR: could not patch CSP permission {}: {}/{}'.format(full_csp_permission_name, 
                                                                                        patch_resp.status_code, 
                                                                                        patch_resp.text))
        else:
            print('Successfully updated permission {}'.format(full_csp_permission_name))

    def delete_permission(self, permission_urn):
        """Delete a permission definition"""
        delete_resp = self.session.delete('/access/v2/permissions/{}'.format(self.permission_urn))
        if not delete_resp.ok:
            raise RuntimeError('ERROR: could not delete permission {}: {}/{}'
                               .format(permission_urn, delete_resp.status_code, delete_resp.text))
        else:
            print('Successfully deleted permission {}'.format(permission_urn))

    def delete_permission_csp(self, permission_name):
        """Delete a CSP permission"""
        retry = True
        while retry:
            delete_resp = csp_session.delete('/csp/gateway/iam-roles-mgmt/api/services/{}/permissions/{}'.format(self.csp_session.service_definition_id,
                                                                                                             permission_name))
            if delete_resp.status_code == 429:
                time.sleep(60)
            else:
                retry = False

        if not delete_resp.ok:
            raise RuntimeError('ERROR: could not delete permission {}: {}/{}'
                               .format(permission_name, delete_resp.status_code, delete_resp.text))
        else:
            print('Successfully deleted permission {}'.format(permission_name))

    def add_role(self, role, org_key=None):
        """Create a new role definition"""
        role_urn = role['role_definition']['urn']
        role_definition = role['role_definition']

        org_api = self.get_org_api(org_key)

        # Add the role definition first
        post_resp = self.session.post('/access/v2{}/roles'.format(org_api), data=json.dumps(role_definition))
        if not post_resp.ok:
            raise RuntimeError('ERROR: could not add role definition {}: {}/{}'
                               .format(role_urn, post_resp.status_code, post_resp.text))
        else:
            print('Successfully added role definition {}'.format(role_urn))

        self.wait_for_consistency('/access/v2{}/roles/{}'.format(org_api, role_urn))

    def update_role(self, role, org_key=None):
        """Update an existing role definition"""
        role_urn = role['role_definition']['urn']
        role_definition = role['role_definition']

        org_api = self.get_org_api(org_key)

        # Get the existing definition to use the correct version
        get_resp = self.session.get('/access/v2{}/roles/{}'.format(org_api, role_urn))
        if not get_resp.ok:
            raise RuntimeError('ERROR: could not retrieve role {} to update:  {}/{}'
                               .format(role_urn, get_resp.status_code, get_resp.text))
        existing_role = json.loads(get_resp.text)
        role_definition['version'] = existing_role['version']

        put_resp = self.session.put('/access/v2{}/roles/{}'.format(org_api, role_urn), data=json.dumps(role_definition))
        if not put_resp.ok:
            raise RuntimeError('ERROR: could not update role definition {}: {}/{}'
                               .format(role_urn, put_resp.status_code, put_resp.text))
        else:
            print('Successfully updated role definition {}'.format(role_urn))

        self.wait_for_consistency('/access/v2{}/roles/{}'.format(org_api, role_urn))

    def delete_role(self, role_urn, org_key=None):
        """Delete a role definition"""
        org_api = self.get_org_api(org_key)

        existing_resp = self.session.get('/access/v2{}/roles/{}/export'.format(org_api, role_urn))
        existing_json = json.loads(existing_resp.text)
        existing_overrides = existing_json['overrides'] if 'overrides' in existing_json else []
        existing_permissions = existing_json['permission'] if 'permission' in existing_json else []
        existing_subroles = existing_json['subroles'] if 'subroles' in existing_json else []

        # Use the bulk APIs to set permissions and subroles to empty lists
        if len(existing_permissions) > 0:
            post_resp = self.session.post(
                '/access/v2{}/roles/{}/permissions/list'.format(org_api, role_urn),
                data=json.dumps({
                    "permissions": []
                })
            )
            if not post_resp.ok:
                raise RuntimeError('ERROR: could not remove role permissions for role {}: {}/{}'
                                   .format(role_urn, post_resp.status_code, post_resp.text))
            else:
                print('Successfully removed all {} permissions from role {}'
                      .format(len(existing_permissions), role_urn))
        # Org specific roles should never have subroles
        elif len(existing_subroles) > 0 and self.org_key is None:
            post_resp = self.session.post('/access/v2/roles/{}/subroles/list'.format(role_urn),
                data=json.dumps({
                    "subroles": []
                })
            )
            if not post_resp.ok:
                raise RuntimeError('ERROR: could not remove role subroles for role {}: {}/{}'
                                   .format(role_urn, post_resp.status_code, post_resp.text))
            else:
                print('Successfully removed all {} subroles from role {}'.format(len(existing_subroles, role_urn)))

        # delete role overrides individually
        if len(existing_overrides) > 0:
            # delete all existing overrides
            for override in existing_overrides:
                override_urn = override['child_urn']
                delete_resp = self.session.delete(
                    '/access/v2{}/roles/{}/overrides/{}'.format(org_api, role_urn, override_urn))
                if not delete_resp.ok:
                    raise RuntimeError('ERROR: could not delete role override for role {} and override {}: {}/{}'
                                       .format(role_urn, override_urn, delete_resp.status_code, delete_resp.text))
        else:
            print("No role overrides to delete for role {}. Skipping...".format(role_urn))

        # Finally delete the role itself
        print('deleting role {}'.format(role_urn))
        delete_resp = self.session.delete('/access/v2{}/roles/{}'.format(org_api, role_urn))
        if not delete_resp.ok:
            raise RuntimeError('ERROR: could not delete role {}:  {}/{}'
                               .format(role_urn, delete_resp.status_code, delete_resp.text))

    def delete_subrole(self, subrole_urn):
        """Delete a subrole definition"""
        existing_resp = self.session.get('/access/v2/subroles/{}/export'.format(subrole_urn))
        existing_json = json.loads(existing_resp.text)
        existing_permissions = existing_json['permission'] if 'permission' in existing_json else []

        # Use the bulk APIs to set permissions and subroles to empty lists
        if len(existing_permissions) > 0:
            post_resp = self.session.post('/access/v2/subroles/{}/permissions/list'.format(subrole_urn),
                                          data=json.dumps({
                                              "permissions": []
                                          }))
            if not post_resp.ok:
                raise RuntimeError('ERROR: could not remove subrole permissions for subrole {}: {}/{}'
                                   .format(subrole_urn, post_resp.status_code, post_resp.text))
            else:
                print('Successfully removed all permissions from subrole {}'.format(subrole_urn))

        # Finally delete the subrole itself
        print('deleting subrole {}'.format(subrole_urn))
        delete_resp = self.session.delete('/access/v2/subroles/{}'.format(subrole_urn))
        if not delete_resp.ok:
            raise RuntimeError('ERROR: could not delete subrole {}: {}/{}'
                               .format(subrole_urn, delete_resp.status_code, delete_resp.text))

    def update_subrole(self, subrole):
        """Update an existing subrole definition"""
        subrole_urn = subrole['urn']
        subrole_copy = dict(subrole)
        subrole_copy.pop('permissions')

        # Get the existing definition to use the correct version
        get_resp = self.session.get('/access/v2/subroles/{}'.format(subrole_urn))
        if not get_resp.ok:
            raise RuntimeError('ERROR: could not retrieve subrole {} to update: {}/{}'
                               .format(subrole_urn, get_resp.status_code, get_resp.text))
        existing_subrole = json.loads(get_resp.text)
        subrole_copy['version'] = existing_subrole['version']

        # Add the subrole definition first
        put_resp = self.session.put('/access/v2/subroles/{}'.format(subrole_urn), data=json.dumps(subrole_copy))
        if not put_resp.ok:
            raise RuntimeError('ERROR: could not update subrole definition {}: {}/{}'
                               .format(subrole_urn, put_resp.status_code, put_resp.text))
        else:
            print('Successfully updated subrole definition {}'.format(subrole_urn))

        self.wait_for_consistency('/access/v2/subroles/{}'.format(subrole_urn))

    def update_subrole_csp(self, subrole):
        """Update CSP subrole"""
        full_csp_subrole_name = "{}:{}".format(self.csp_session.service_ticker, subrole['urn'])
        api = '/csp/gateway/iam-roles-mgmt/api/services/{}/permissions/{}'.format(self.csp_session.service_definition_id, full_csp_subrole_name)
        patch_payload = { "displayName": subrole['displayName'], "description": subrole['description'] }
        retry = True
        while retry:
            patch_resp = self.csp_session.patch(api, data=json.dumps(patch_payload))
            if patch_resp.status_code == 429:
                time.sleep(60)
            else:
                retry = False

        if not patch_resp.ok:
            raise RuntimeError('ERROR: could not patch subrole {}: {}/{}'.format(full_csp_subrole_name, 
                                                                                 patch_resp.status_code, 
                                                                                 patch_resp.text))
        else:
            print('Successfully updated permission {}'.format(subrole['urn']))

    def add_subrole(self, subrole):
        """Create a new subrole definition"""
        subrole_urn = subrole['urn']
        subrole_copy = dict(subrole)
        subrole_copy.pop('permissions')

        # Add the subrole definition first
        post_resp = self.session.post('/access/v2/subroles', data=json.dumps(subrole_copy))
        if not post_resp.ok:
            raise RuntimeError('ERROR: could not add subrole definition {}: {}/{}'
                               .format(subrole_urn, post_resp.status_code, post_resp.text))
        else:
            print('Successfully added subrole definition {}'.format(subrole_urn))

        self.wait_for_consistency('/access/v2/subroles/{}'.format(subrole_urn))

    def add_subrole_csp(self, subrole):
        """In this context what we refer to subroles are actually CSP permissions"""
        """The name of the permission gets the serviceTicker prepended, once POSTed to CSP"""
        csp_subrole_name = subrole['urn']
        csp_payload = { "name": csp_subrole_name, 
                        "displayName": subrole['displayName'], 
                        "description": subrole['description'], 
                        "customRoleSupport": "true" }
        retry = True
        while retry:
            post_resp = self.csp_session.post('/csp/gateway/iam-roles-mgmt/api/services/{}/permissions'.format(self.csp_session.service_definition_id), 
                                                                                                           data=json.dumps(csp_payload))
            if post_resp.status_code == 429:
                time.sleep(60)
            else:
                retry = False

        if not post_resp.ok:
            raise RuntimeError('ERROR: could not add CSP permission {}: {}/{}'.format(csp_subrole_name, post_resp.status_code, post_resp.text))
        else:
            print("Successfully added permission {} to CSP".format(csp_subrole_name))

    def update_role_permissions(self, role, org_key=None):
        """Update all role permissions of a role"""
        role_urn = role['role_definition']['urn']
        role_permissions = role['permissions'] if ('permissions' in role and role['permissions']) else []
        if len(role_permissions) == 0:
            print('Skipping permissions update for role {}'.format(role_urn))
            return

        org_api = self.get_org_api(org_key)
        post_resp = self.session.post('/access/v2{}/roles/{}/permissions/list'.format(org_api, role_urn),
                                      data=json.dumps({
                                          "permissions": role_permissions
                                      }))
        if not post_resp.ok:
            raise RuntimeError('ERROR: could not add role permissions for role {}: {}/{}'
                               .format(role_urn, post_resp.status_code, post_resp.text))
        else:
            print('Successfully updated permissions of role {}'.format(role_urn))

    def invalidate_role_cache(self, role):
        """Invalidate the cache entry for a given role"""
        role_urn = role['role_definition']['urn']

        delete_resp = self.session.delete('/access/v2/caches/roles/{}/permissions'.format(role_urn))

        if not delete_resp.ok:
            raise RuntimeError('ERROR: could not invalidate cache for role {}: {}/{}'
                               .format(role_urn, delete_resp.status_code, delete_resp.text))
        else:
            print('Successfully invalidated cache for role {}'.format(role_urn))

    def update_role_subroles(self, role, org_key=None):
        """Update all role subroles of a role"""
        role_urn = role['role_definition']['urn']
        role_subroles = role['subroles'] if ('subroles' in role and role['subroles'] is not None) else []
        if len(role_subroles) == 0:
            print('Skipping subrole update for role {}'.format(role_urn))
            return

        org_api = self.get_org_api(org_key)
        post_resp = self.session.post('/access/v2{}/roles/{}/subroles/list'.format(org_api, role_urn),
                                      data=json.dumps({
                                          "subroles": role_subroles
                                      }))
        if not post_resp.ok:
            raise RuntimeError('ERROR: could not add role subroles for role {}: {}/{}'
                               .format(role_urn, post_resp.status_code, post_resp.text))
        else:
            print('Successfully updated subroles of role {}'.format(role_urn))

    def update_role_overrides(self, role, org_key=None):
        """Update all role overrides of a role
        deleting overrides is not implemented"""
        role_urn = role['role_definition']['urn']
        role_overrides = role['overrides'] if ('overrides' in role and role['overrides'] is not None) else []

        org_api = self.get_org_api(org_key)
        override_success = True
        for override in role_overrides:
            override_urn = override['child_urn']
            get_resp = self.session.get('/access/v2{}/roles/{}/overrides/{}'
                                        .format(org_api, role_urn, override_urn))

            if get_resp.ok:
                # Override exists and should be updated
                existing_override = json.loads(get_resp.text)
                override['version'] = existing_override['version']

                put_resp = self.session.put(
                    '/access/v2{}/roles/{}/overrides/{}'.format(org_api, role_urn, override_urn),
                        data=json.dumps(override)
                    )
                if not put_resp.ok:
                    override_success = False
                    raise RuntimeError("ERROR: could not update role override for role {} and override {}: "
                                       "PUT request failed: {}/{}".format(role_urn, override['child_urn'],
                                                                          put_resp.status_code,
                                                                          put_resp.text))
                else:
                    print('-successfully updated override {} for role {}'.format(override_urn, role_urn))
            elif get_resp.status_code == 404:
                post_resp = self.session.post('/access/v2{}/roles/{}/overrides/{}'
                                              .format(org_api, role_urn, override_urn),
                    data=json.dumps(override)
                )
                if not post_resp.ok:
                    raise RuntimeError("ERROR: could not update role override for role {} and override {}: "
                                       "POST request failed: {}/{}".format(role_urn, override['child_urn'],
                                                                           post_resp.status_code,
                                                                           post_resp.text))
                else:
                    print('-successfully added override {} for role {}'.format(override_urn, role_urn))
            else:
                raise RuntimeError("ERROR: could not get role override for role {} and override {}:  " \
                      "GET request failed: {}/{}".format(role_urn, override['child_urn'],
                                                         put_resp.status_code,
                                                         put_resp.text))
        if override_success:
            print('successfully updated all overrides for role {}'.format(role_urn))

    def update_subrole_permissions(self, subrole):
        """Update all permissions of a subrole"""
        subrole_urn = subrole['urn']
        subrole_permissions = subrole['permissions']
        post_resp = self.session.post(
            '/access/v2/subroles/{}/permissions/list'.format(subrole_urn),
            data=json.dumps({
                "permissions": subrole_permissions
            })
        )
        if not post_resp.ok:
            raise RuntimeError('ERROR: could not add subrole permissions for subrole {}: Status: {}, Message: {}'
                               .format(subrole_urn, post_resp.status_code, post_resp.text))
        else:
            print('Successfully updated permissions of subrole {}'.format)

    def get_org_api(self, org_key=None):
        """Get the string used for org APIs, or an empty string if there is no org key"""
        org_api = ''
        if org_key is not None:
            org_api = '/orgs/{}'.format(org_key)
        return org_api

    def wait_for_consistency(self, api):
        """Make a GET request to an API until either a 200 or a non-404 response code is returned"""
        added = False
        while not added:
            resp = self.session.get(api)
            if resp.ok:
                added = True
            elif resp.status_code == 404:
                print(".")
                time.sleep(0.25)
            else:
                raise RuntimeError('Unexpected response while waiting for DB consistency for API {}: {}/{}'
                                   .format(api, resp.status_code, resp.text))

