import json

import requests


class RequestsSession(requests.Session):
    """Wrapper class for requests.Session which uses a base URL instance variable"""

    def __init__(self, base_url=None, *args, **kwargs):
        super(RequestsSession, self).__init__(*args, **kwargs)
        self.base_url = base_url

    def get(self, path, **kwargs):
        full_url = '{}{}'.format(self.base_url, path)
        return super(RequestsSession, self).get(full_url, **kwargs)

    def delete(self, path, **kwargs):
        full_url = '{}{}'.format(self.base_url, path)
        return super(RequestsSession, self).delete(full_url, **kwargs)

    def post(self, path, data=None, json=None, **kwargs):
        full_url = '{}{}'.format(self.base_url, path)
        return super(RequestsSession, self).post(full_url, data, json, **kwargs)

    def put(self, path, data=None, **kwargs):
        full_url = '{}{}'.format(self.base_url, path)
        return super(RequestsSession, self).put(full_url, data, **kwargs)

    def patch(self, path, data=None, **kwargs):
        full_url = '{}{}'.format(self.base_url, path)
        return super(RequestsSession, self).patch(full_url, data, **kwargs)


class CSPSession(RequestsSession):
    def __init__(self, base_url=None, oauth_credentials=None, service_definition_id=None, *args, **kwargs):
        super(CSPSession, self).__init__(base_url=base_url, *args, **kwargs)
        try:
            self.access_token = self.__get_access_token(oauth_credentials)
        except:
            raise RuntimeError("CSP access token could not be acquired")
        self.headers.update({ "Authorization": "Bearer " + self.access_token })
        self.service_definition_id = service_definition_id
        try:
            self.service_ticker = self.__get_service_ticker()
        except:
            raise RuntimeError("Service ticker could not be acquired from CSP")

    def __get_access_token(self, oauth_credentials):
        r = super().post("/csp/gateway/am/api/auth/token",
                        headers={ "Content-Type": "application/x-www-form-urlencoded",
                                  "Authorization" : "Basic " + oauth_credentials },
                        data={ "grant_type": "client_credentials" }, allow_redirects=True)
        if not r.ok:
            print(r.text)

        return json.loads(r.text)["access_token"]

    def __get_service_ticker(self):
        r = super().get("/csp/gateway/slc/api/definitions/external/{}".format(self.service_definition_id),
                        allow_redirects=True)

        return json.loads(r.text)["serviceTicker"]


class MigrationSession(RequestsSession):
    def __init__(self, base_url=None, green_only=False, *args, **kwargs):
        super(MigrationSession, self).__init__(base_url=base_url, *args, **kwargs)
        self.base_url = base_url
        self.params['modify_uuid'] = False
        self.params['access_green_only'] = green_only
